import re
from typing import Final

from aioairtable.helpers import get_python_version, get_software

VERSION_RXP_STR: Final[str] = r"\d\.\d{1,2}\.[a-z0-9]+"


def test_get_python_version() -> None:
    version_match = re.match(VERSION_RXP_STR, get_python_version())
    assert version_match is not None


def test_get_software() -> None:
    software_rxp = f"Python/{VERSION_RXP_STR} aioairtable/{VERSION_RXP_STR}"
    assert re.match(software_rxp, get_software()) is not None



import asyncio
from datetime import UTC, datetime
from typing import AsyncGenerator, Final
from unittest.mock import call

import msgspec.json
import pytest
import pytest_asyncio
from msgspec import Struct
from pytest_mock import MockerFixture
from yarl import URL

from aioairtable import Airtable
from aioairtable import aioairtable as aat
from aioairtable.aioairtable import (
    AirtableRecord,
    CellFormat,
    DeletedRecord,
    Record,
    RecordList,
    RecordRequest,
    SortDirection,
)

DT_FORMAT: Final = "%Y-%m-%dT%H:%M:%S.000Z"


def parse_dt(string: str) -> datetime:
    return datetime.strptime(string, DT_FORMAT).replace(tzinfo=UTC)


@pytest.fixture
def dt_str() -> str:
    return datetime.now().strftime(DT_FORMAT)


@pytest.fixture
def url() -> URL:
    return URL("https://example.com")


@pytest_asyncio.fixture
async def _airtable() -> AsyncGenerator[Airtable, None]:
    airtable = Airtable("secret_key")
    yield airtable
    await airtable.close()


class SomeResponseData(Struct, frozen=True):
    some_key: int


@pytest.fixture
def some_response_data() -> SomeResponseData:
    return SomeResponseData(55)


@pytest.mark.asyncio
async def test_airtable_underscore_request(
    _airtable: Airtable,
    url: URL,
    some_response_data: SomeResponseData,
    mocker: MockerFixture,
) -> None:
    request = mocker.patch.object(_airtable._client, "request")
    response = request.return_value.__aenter__.return_value
    response.read.return_value = msgspec.json.encode(some_response_data)
    assert (
        await _airtable._request("GET", url, SomeResponseData)
        == some_response_data
    )
    request.assert_called_once_with(
        "GET",
        url,
        headers={"Authorization": "Bearer secret_key"},
        data=None,
    )
    response.read.assert_awaited_once_with()


@pytest.mark.asyncio
async def test_airtable_request(
    _airtable: Airtable,
    url: URL,
    some_response_data: SomeResponseData,
    mocker: MockerFixture,
) -> None:
    loop = asyncio.get_running_loop()
    request = mocker.patch.object(_airtable, "_request")
    request.return_value = some_response_data
    assert (
        await _airtable.request(
            "some_base_id",
            "GET",
            url,
            SomeResponseData,
        )
        == some_response_data
    )
    request.assert_awaited_once_with("GET", url, SomeResponseData, None)
    time1 = loop.time()
    assert (
        await _airtable.request(
            "some_base_id",
            "GET",
            url,
            SomeResponseData,
        )
        == some_response_data
    )
    time2 = loop.time()
    assert time2 - time1 >= aat.AT_INTERVAL


@pytest.mark.asyncio
async def test_airtable_base_request(
    _airtable: Airtable,
    url: URL,
    some_response_data: SomeResponseData,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    request = mocker.patch.object(base._airtable, "request")
    request.return_value = some_response_data
    assert (
        await base.request(
            "GET",
            url,
            SomeResponseData,
        )
        == some_response_data
    )
    request.assert_awaited_once_with(
        "some_base_id",
        "GET",
        url,
        SomeResponseData,
        None,
    )


@pytest.mark.asyncio
async def test_airtable_table_request(
    _airtable: Airtable,
    url: URL,
    some_response_data: SomeResponseData,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    request = mocker.patch.object(table._base, "request")
    request.return_value = some_response_data
    assert (
        await table._request(
            "GET",
            url,
            SomeResponseData,
        )
        == some_response_data
    )
    request.assert_awaited_once_with(
        "GET",
        url,
        SomeResponseData,
        None,
    )


@pytest.mark.asyncio
async def test_airtable_table_list_records(
    _airtable: Airtable,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    request = mocker.patch.object(table, "_request")
    request.return_value = RecordList(records=())
    records, offset = await table.list_records(
        fields=("field1", "field2", "field3"),
        filter_by_formula="{field4}",
        max_records=100500,
        page_size=10,
        sort=(("field5", SortDirection.ASC), ("field6", SortDirection.DESC)),
        view="table3",
        cell_format=CellFormat.JSON,
        time_zone="Europe/Moscow",
        user_locale="ru",
        offset="offset22",
    )
    request.assert_awaited_once_with(
        "GET",
        table.url.with_query(
            (
                ("fields[]", "field1"),
                ("fields[]", "field2"),
                ("fields[]", "field3"),
                ("filterByFormula", "{field4}"),
                ("maxRecords", "100500"),
                ("pageSize", "10"),
                ("sort[0][field]", "field5"),
                ("sort[0][direction]", "asc"),
                ("sort[1][field]", "field6"),
                ("sort[1][direction]", "desc"),
                ("view", "table3"),
                ("cellFormat", "json"),
                ("timeZone", "Europe/Moscow"),
                ("userLocale", "ru"),
                ("offset", "offset22"),
            )
        ),
        type_=RecordList[Struct],
    )
    assert records == ()
    assert offset is None


@pytest.mark.asyncio
async def test_airtable_table_iter_records(
    _airtable: Airtable,
    dt_str: str,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    request = mocker.patch.object(table, "_request")
    now = datetime.now(UTC)
    request.side_effect = (
        RecordList(
            records=(
                Record(id="record1", fields=Struct(), created_time=now),
                Record(id="record2", fields=Struct(), created_time=now),
                Record(id="record3", fields=Struct(), created_time=now),
            ),
            offset="offset1",
        ),
        RecordList(
            records=(
                Record(id="record4", fields=Struct(), created_time=now),
                Record(id="record5", fields=Struct(), created_time=now),
                Record(id="record6", fields=Struct(), created_time=now),
            ),
        ),
    )
    records = tuple([record async for record in table.iter_records()])
    assert request.await_args_list == [
        call(
            "GET",
            table.url.with_query(pageSize=25),
            type_=RecordList[Struct],
        ),
        call(
            "GET",
            table.url.with_query(pageSize=25, offset="offset1"),
            type_=RecordList[Struct],
        ),
    ]
    assert all(isinstance(record, AirtableRecord) for record in records)
    record_ids = (
        "record1",
        "record2",
        "record3",
        "record4",
        "record5",
        "record6",
    )
    assert tuple(record.id for record in records) == record_ids


@pytest.mark.asyncio
async def test_airtable_table_retrieve_record(
    _airtable: Airtable,
    dt_str: str,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    request = mocker.patch.object(table, "_request")
    now = datetime.now(UTC)
    record = Record(
        id="record1",
        fields=Struct(),
        created_time=now,
    )
    request.return_value = record
    at_record = await table.retrieve_record("record1")
    request.assert_awaited_once_with(
        "GET",
        table.url / "record1",
        type_=Record[Struct],
    )
    assert isinstance(at_record, AirtableRecord)
    assert at_record.id == "record1"
    assert at_record.fields == Struct()
    assert at_record.created_time == now
    assert at_record.table == table


@pytest.mark.asyncio
async def test_airtable_table_create_record(
    _airtable: Airtable,
    dt_str: str,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    request = mocker.patch.object(table, "_request")
    now = datetime.now(UTC)
    request.return_value = Record(
        id="record1",
        fields=Struct(),
        created_time=now,
    )
    record = await table.create_record(Struct())
    request.assert_awaited_once_with(
        "POST",
        table.url,
        payload=RecordRequest(Struct()),
        type_=Record[Struct],
    )
    assert isinstance(record, AirtableRecord)
    assert record.id == "record1"
    assert record.fields == Struct()
    assert record.created_time == now
    assert record.table == table


@pytest.mark.asyncio
async def test_airtable_record_request(
    _airtable: Airtable,
    dt_str: str,
    url: URL,
    some_response_data: SomeResponseData,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    record = AirtableRecord(
        "record1",
        Struct(),
        datetime.now(UTC),
        table,
    )
    request = mocker.patch.object(record.table._base, "request")
    request.return_value = some_response_data
    assert (
        await table._request(
            "GET",
            url,
            SomeResponseData,
        )
        == some_response_data
    )
    request.assert_awaited_once_with(
        "GET",
        url,
        SomeResponseData,
        None,
    )


@pytest.mark.asyncio
async def test_airtable_record_update(
    _airtable: Airtable,
    dt_str: str,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    record = AirtableRecord(
        "record1",
        Struct(),
        datetime.now(UTC),
        table,
    )
    request = mocker.patch.object(record, "_request")
    request.return_value = Record(
        "record1",
        Struct(),
        datetime.now(UTC),
    )
    await record.update(
        Struct(),
    )
    assert record.fields == Struct()
    request.assert_awaited_once_with(
        "PATCH",
        record.url,
        type_=Record[Struct],
        payload=RecordRequest(Struct()),
    )


@pytest.mark.asyncio
async def test_airtable_record_delete(
    _airtable: Airtable,
    dt_str: str,
    mocker: MockerFixture,
) -> None:
    base = _airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    record = AirtableRecord(
        "record1",
        Struct(),
        datetime.now(UTC),
        table,
    )
    request = mocker.patch.object(record, "_request")
    request.return_value = DeletedRecord("record1", True)
    await record.delete()
    assert record.deleted
    request.assert_awaited_once_with(
        "DELETE",
        record.url,
        type_=DeletedRecord,
    )



import logging
from datetime import datetime
from enum import IntEnum, StrEnum, unique
from types import TracebackType
from typing import (
    Any,
    AsyncIterator,
    Final,
    Generator,
    Generic,
    Iterable,
    Literal,
    Optional,
    Self,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import backoff
import msgspec.json
from aiofreqlimit import FreqLimit
from aiohttp import BaseConnector, ClientResponseError, ClientSession
from msgspec import Struct, field
from multidict import MultiDict
from yarl import URL

from .helpers import get_software

__all__ = (
    "Airtable",
    "AirtableBase",
    "AirtableRecord",
    "AirtableTable",
    "Attachment",
    "CellFormat",
    "Collaborator",
    "Method",
    "NewAttachment",
    "RecordList",
    "SortDirection",
    "Thumbnail",
)

SOFTWARE: Final = get_software()
API_URL: Final = URL("https://api.airtable.com/v0")
AT_INTERVAL: Final = 1 / 5
AT_WAIT: Final = 30.0


@unique
class BackoffCodes(IntEnum):
    TOO_MANY_REQUESTS = 429
    BAD_GATEWAY = 502
    SERVICE_UNAVAILABLE = 503
    GATEWAY_TIMEOUT = 504


Method = Literal["GET", "POST", "PATCH", "DELETE"]

logger = logging.getLogger("airtable")


_T = TypeVar("_T", bound=Struct)


class Record(Struct, Generic[_T], frozen=True):
    id: str
    fields: _T
    created_time: datetime = field(name="createdTime")


class RecordRequest(Struct, Generic[_T], frozen=True):
    fields: _T


class DeletedRecord(Struct, frozen=True):
    id: str
    deleted: bool


class RecordList(Struct, Generic[_T], frozen=True):
    records: tuple[Record[_T], ...]
    offset: str | None = None


class Thumbnail(Struct, frozen=True):
    url: str
    width: int
    height: int


class NewAttachment(Struct, frozen=True, omit_defaults=True):
    url: str
    id: str | None = None
    filename: str | None = None


class Attachment(Struct, frozen=True, omit_defaults=True):
    id: str
    url: str
    filename: str
    size: int
    type: str
    width: int | None = None
    height: int | None = None
    thumbnails: tuple[Thumbnail, ...] | None = None


class Collaborator(Struct, frozen=True):
    id: str
    email: str
    name: str


@unique
class SortDirection(StrEnum):
    ASC = "asc"
    DESC = "desc"


@unique
class CellFormat(StrEnum):
    JSON = "json"
    STRING = "string"


def backoff_wait_gen(at_wait: float) -> Generator[float, Any, None]:
    expo_gen = backoff.expo()
    yield expo_gen.send(None)
    for value in expo_gen:
        yield at_wait + value


def backoff_giveup(exception: Exception) -> bool:
    assert isinstance(exception, ClientResponseError)
    try:
        BackoffCodes(exception.status)
    except ValueError:
        return True
    else:
        return False


def build_repr(class_name: str, **kwargs: Any) -> str:
    args = ", ".join(f"{key}={value!r}" for key, value in kwargs.items())
    return f"{class_name}({args})"


_R = TypeVar("_R", bound=Struct)


class Airtable:
    def __init__(
        self,
        api_key: str,
        connector: BaseConnector | None = None,
    ) -> None:
        self._auth_headers = {"Authorization": f"Bearer {api_key}"}
        self._client = ClientSession(
            connector=connector,
            headers={"User-Agent": SOFTWARE},
            raise_for_status=True,
        )
        self._freq_limit = FreqLimit(AT_INTERVAL)

    def __repr__(self) -> str:
        return build_repr("Airtable", api_key="...")

    @property
    def client(self) -> ClientSession:
        return self._client

    async def _request(
        self,
        method: Method,
        url: URL,
        type_: Type[_R],
        payload: Struct | None = None,
    ) -> _R:
        async with self._client.request(
            method,
            url,
            headers=self._auth_headers,
            data=msgspec.json.encode(payload) if payload is not None else None,
        ) as client_response:
            logger.debug(
                "Request %s %s %r",
                method,
                url.human_repr(),
                payload,
            )
            response_data = await client_response.read()
            response = msgspec.json.decode(response_data, type=type_)
            logger.debug(
                "Response %r",
                response,
            )
            return response

    @backoff.on_exception(
        backoff_wait_gen,
        ClientResponseError,
        giveup=backoff_giveup,
        at_wait=AT_WAIT,
    )
    async def request(
        self,
        base_id: str,
        method: Method,
        url: URL,
        type_: Type[_R],
        payload: Struct | None = None,
    ) -> _R:
        async with self._freq_limit.resource(base_id):
            return await self._request(
                method,
                url,
                type_,
                payload,
            )

    async def close(self) -> None:
        await self._client.close()
        await self._freq_limit.clear()

    def base(self, base_id: str) -> "AirtableBase":
        return AirtableBase(base_id, self)

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.close()


class AirtableBase:
    def __init__(
        self,
        base_id: str,
        airtable: Airtable,
    ) -> None:
        self._airtable: Final[Airtable] = airtable
        self._id: Final[str] = base_id
        self._url: Final[URL] = API_URL / base_id

    def __repr__(self) -> str:
        return build_repr(
            "AirtableBase",
            base_id=self._id,
            airtable=self._airtable,
        )

    @property
    def id(self) -> str:
        return self._id

    @property
    def url(self) -> URL:
        return self._url

    async def request(
        self,
        method: Method,
        url: URL,
        type_: Type[_R],
        payload: Any = None,
    ) -> _R:
        return await self._airtable.request(
            self._id,
            method,
            url,
            type_,
            payload,
        )

    def table(
        self, table_name: str, fields_type: Type[_T]
    ) -> "AirtableTable[_T]":
        return AirtableTable(table_name, self, fields_type)


class AirtableTable(Generic[_T]):
    def __init__(
        self,
        table_name: str,
        base: AirtableBase,
        fields_type: Type[_T],
    ) -> None:
        self._name: Final = table_name
        self._base: Final = base
        self._url: Final[URL] = base.url / table_name
        self._fields_type: Final = fields_type

    def __repr__(self) -> str:
        return build_repr(
            "AirtableTable", table_name=self._name, base=self._base
        )

    @property
    def name(self) -> str:
        return self._name

    @property
    def url(self) -> URL:
        return self._url

    @property
    def base(self) -> AirtableBase:
        return self._base

    @property
    def fields_type(self) -> Type[_T]:
        return self._fields_type

    async def _request(
        self,
        method: Method,
        url: URL,
        type_: Type[_R],
        payload: Any = None,
    ) -> _R:
        return await self._base.request(
            method,
            url,
            type_,
            payload,
        )

    async def list_records(
        self,
        *,
        fields: Iterable[str] | None = None,
        filter_by_formula: str | None = None,
        max_records: int | None = None,
        page_size: int | None = None,
        sort: Iterable[tuple[str, SortDirection]] | None = None,
        view: str | None = None,
        cell_format: CellFormat | None = None,
        time_zone: str | None = None,
        user_locale: str | None = None,
        offset: str | None = None,
    ) -> tuple[tuple["AirtableRecord[_T]", ...], str | None]:
        params: MultiDict[Union[int, str]] = MultiDict()
        if fields is not None:
            params.extend(("fields[]", fields) for fields in fields)
        if filter_by_formula is not None:
            params.add("filterByFormula", filter_by_formula)
        if max_records is not None:
            params.add("maxRecords", max_records)
        if page_size is not None:
            params.add("pageSize", page_size)
        if sort is not None:
            for index, (field_name, direction) in enumerate(sort):
                params.add(f"sort[{index}][field]", field_name)
                params.add(f"sort[{index}][direction]", direction)
        if view is not None:
            params.add("view", view)
        if cell_format is not None:
            params.add("cellFormat", cell_format)
        if time_zone is not None:
            params.add("timeZone", time_zone)
        if user_locale is not None:
            params.add("userLocale", user_locale)
        if offset is not None:
            params.add("offset", offset)
        url = self._url.with_query(params)
        fields_type = self._fields_type
        record_list = await self._request(
            "GET",
            url,
            type_=RecordList[fields_type],  # type: ignore[valid-type]
        )
        records = tuple(
            AirtableRecord(
                record.id,
                record.fields,
                record.created_time,
                self,
            )
            for record in record_list.records
        )
        return records, record_list.offset

    async def iter_records(
        self,
        *,
        fields: Optional[Iterable[str]] = None,
        filter_by_formula: Optional[str] = None,
        max_records: Optional[int] = None,
        page_size: int = 25,
        sort: Optional[Iterable[Tuple[str, SortDirection]]] = None,
        view: Optional[str] = None,
        cell_format: Optional[CellFormat] = None,
        time_zone: Optional[str] = None,
        user_locale: Optional[str] = None,
    ) -> AsyncIterator["AirtableRecord[_T]"]:
        offset: Optional[str] = None
        while True:
            records, offset = await self.list_records(
                fields=fields,
                filter_by_formula=filter_by_formula,
                max_records=max_records,
                page_size=page_size,
                sort=sort,
                view=view,
                cell_format=cell_format,
                time_zone=time_zone,
                user_locale=user_locale,
                offset=offset,
            )
            for record in records:
                yield record
            if offset is None:
                break

    async def retrieve_record(
        self,
        record_id: str,
    ) -> "AirtableRecord[_T]":
        fields_type = self._fields_type
        record = await self._request(
            "GET",
            self._url / record_id,
            type_=Record[fields_type],  # type: ignore[valid-type]
        )
        return AirtableRecord(
            record.id,
            record.fields,
            record.created_time,
            self,
        )

    async def create_record(
        self,
        fields: _T,
    ) -> "AirtableRecord[_T]":
        fields_type = self._fields_type
        record = await self._request(
            "POST",
            self._url,
            type_=Record[fields_type],  # type: ignore[valid-type]
            payload=RecordRequest(fields),
        )
        return AirtableRecord(
            record.id,
            record.fields,
            record.created_time,
            table=self,
        )


class AirtableRecord(Generic[_T]):
    def __init__(
        self,
        record_id: str,
        fields: _T,
        created_time: datetime,
        table: AirtableTable[_T],
    ) -> None:
        self._table: Final = table
        self._id: Final = record_id
        self._url: Final = table.url / record_id
        self._fields = fields
        self._created_time: Final = created_time
        self._deleted: bool = False

    def __repr__(self) -> str:
        return build_repr(
            "AirtableRecord",
            record_id=self._id,
            fields=self._fields,
            created_time=self._created_time,
            table=self._table,
        )

    @property
    def id(self) -> str:
        return self._id

    @property
    def url(self) -> URL:
        return self._url

    @property
    def fields(self) -> _T:
        return self._fields

    @property
    def created_time(self) -> datetime:
        return self._created_time

    @property
    def table(self) -> AirtableTable[_T]:
        return self._table

    @property
    def deleted(self) -> bool:
        return self._deleted

    async def _request(
        self,
        method: Method,
        url: URL,
        type_: Type[_R],
        payload: Struct | None = None,
    ) -> _R:
        return await self._table.base.request(
            method,
            url,
            type_,
            payload,
        )

    async def update(
        self,
        fields: _T,
    ) -> None:
        if self._deleted:
            raise RuntimeError("Record is deleted")
        fields_type = self._table.fields_type
        record = await self._request(
            "PATCH",
            self._url,
            payload=RecordRequest(fields),
            type_=Record[fields_type],  # type: ignore[valid-type]
        )
        self._fields = record.fields

    async def delete(self) -> None:
        if self._deleted:
            raise RuntimeError("Record is already deleted")
        record = await self._request(
            "DELETE",
            self._url,
            type_=DeletedRecord,
        )
        assert record.deleted
        self._deleted = record.deleted



__all__ = ("get_python_version", "get_software")


def get_python_version() -> str:
    from sys import version_info as version

    return f"{version.major}.{version.minor}.{version.micro}"


def get_software() -> str:
    from . import __version__

    return f"Python/{get_python_version()} aioairtable/{__version__}"



import asyncio
import os
import re
from dataclasses import dataclass
from datetime import UTC, datetime, timezone
from tempfile import mkdtemp
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Final,
    Mapping,
    Optional,
    Protocol,
    TypedDict,
    cast,
    runtime_checkable,
)

import backoff
import pytest
import pytest_asyncio
from aiohttp import (
    ClientResponseError,
    ClientSession,
    RequestInfo,
    UnixConnector,
)
from aiohttp.web import (
    Application,
    AppRunner,
    HTTPBadRequest,
    HTTPNotFound,
    Request,
    Response,
    StreamResponse,
    UnixSite,
    delete,
    get,
    json_response,
    middleware,
    patch,
    post,
)
from hypothesis import given
from hypothesis.strategies import integers
from msgspec import Struct
from multidict import CIMultiDict, CIMultiDictProxy
from yarl import URL

from aioairtable import aioairtable as aat
from aioairtable.aioairtable import (
    Airtable,
    AirtableBase,
    AirtableRecord,
    AirtableTable,
    CellFormat,
    RecordList,
    SortDirection,
    backoff_giveup,
    backoff_wait_gen,
)

DT_FORMAT: Final = "%Y-%m-%dT%H:%M:%S.000Z"


def parse_dt(string: str) -> datetime:
    return datetime.strptime(string, DT_FORMAT).replace(tzinfo=timezone.utc)


Fields = Mapping[str, Any]


class Record(TypedDict):
    id: str
    fields: Fields
    createdTime: str


class DeletedRecord(TypedDict):
    id: str
    deleted: bool


@dataclass(frozen=True)
class RequestData:
    method: str
    url: URL
    data: Any


Handler = Callable[[Request], Awaitable[StreamResponse]]


@runtime_checkable
class SupportsLessThan(Protocol):
    def __lt__(self, other: Any) -> bool: ...


def fields_sort_key(field: str) -> Callable[[Record], SupportsLessThan]:
    def _fields_sort_key(record: Record) -> SupportsLessThan:
        if not isinstance(record["fields"][field], SupportsLessThan):
            raise HTTPBadRequest(reason=f'"{field}" type not sortable')
        return cast(SupportsLessThan, record["fields"][field])

    return _fields_sort_key


class AirtableServer:
    def __init__(self, api_key: str) -> None:
        self._started: bool = False
        self._tmp_dir: str | None = None
        self._connector: UnixConnector | None = None
        self._loop = asyncio.get_running_loop()
        self._tables: dict[tuple[str, str], list[Record]] = {}
        self._requests: list[RequestData] = []
        self._api_key: Final = api_key
        application = Application(middlewares=(self._auth, self._log))
        application.router.add_routes(
            (
                get(
                    "/v0/{base_id}/{table_name}",
                    self.list_records,
                ),
                get(
                    "/v0/{base_id}/{table_name}/{record_id}",
                    self.retrieve_record,
                ),
                post(
                    "/v0/{base_id}/{table_name}",
                    self.create_record,
                ),
                patch(
                    "/v0/{base_id}/{table_name}/{record_id}",
                    self.update_record,
                ),
                delete(
                    "/v0/{base_id}/{table_name}/{record_id}",
                    self.delete_record,
                ),
            )
        )
        self._runner = AppRunner(application)

    @middleware
    async def _auth(
        self,
        request: Request,
        handler: Handler,
    ) -> StreamResponse:
        if "Authorization" not in request.headers:
            raise HTTPBadRequest(reason="Authorization header absent")
        if request.headers["Authorization"] != f"Bearer {self._api_key}":
            raise HTTPBadRequest(reason="Wrong authorization header")
        if request.headers.get("User-Agent") != aat.SOFTWARE:
            raise HTTPBadRequest(reason="Wrong user-agent header")
        return await handler(request)

    @middleware
    async def _log(
        self,
        request: Request,
        handler: Handler,
    ) -> StreamResponse:
        url = request.url.with_scheme("https")
        has_data = request.method in ("POST", "PATCH")
        data = await request.json() if has_data else None
        self._requests.append(RequestData(request.method, url, data))
        return await handler(request)

    def requests(self) -> list[RequestData]:
        return self._requests.copy()

    def add_records(
        self,
        base_id: str,
        table_name: str,
        records: list[Record],
    ) -> None:
        if (base_id, table_name) in self._tables:
            fields_keys = self._tables[base_id, table_name][0]["fields"].keys()
            assert all(
                record["fields"].keys() == fields_keys for record in records
            )
            self._tables[base_id, table_name].extend(records)
        else:
            self._tables[base_id, table_name] = records

    async def start(self) -> None:
        if self._started:
            raise RuntimeError("Server already started")
        self._tmp_dir = await self._loop.run_in_executor(None, mkdtemp)
        await self._runner.setup()
        await UnixSite(self._runner, self._socket_path).start()
        self._started = True

    async def stop(self) -> None:
        if not self._started:
            raise RuntimeError("Server not started")
        assert self._tmp_dir is not None
        await self._runner.cleanup()
        await self._loop.run_in_executor(None, os.remove, self._socket_path)
        await self._loop.run_in_executor(None, os.rmdir, self._tmp_dir)
        self._started = False

    @property
    def _socket_path(self) -> str:
        return f"{self._tmp_dir}/site.socket"

    @property
    def connector(self) -> UnixConnector:
        if self._connector is None:
            self._connector = UnixConnector(self._socket_path)
        return self._connector

    async def list_records(self, request: Request) -> Response:
        base_id = request.match_info["base_id"]
        table_name = request.match_info["table_name"]
        if (base_id, table_name) not in self._tables:
            raise HTTPNotFound(reason="Table not found")
        sort: dict[int, dict[str, str]] = {}
        for key in request.query:
            if (
                not key.startswith("fields")
                and len(request.query.getall(key)) > 1
            ):
                raise HTTPBadRequest(reason="Wrong sort parameter1")
            match = re.match(r"sort\[([^]]+)]\[([^]]+)]", key)
            if match is not None:
                index, fd = match.groups()
                if not index.isdigit() or fd not in ("field", "direction"):
                    raise HTTPBadRequest(reason="Wrong sort parameter2")
                if int(index) in sort:
                    sort[int(index)][fd] = request.query[key]
                else:
                    sort[int(index)] = {fd: request.query[key]}
        if not all("field" in item for item in sort.values()):
            raise HTTPBadRequest(reason="Wrong sort parameter3")
        records = self._tables[base_id, table_name].copy()
        for sort_index in sorted(sort, reverse=True):
            field = sort[sort_index]["field"]
            if field not in records[0]["fields"]:
                raise HTTPBadRequest(reason="Wrong sort parameter4")
            direction = sort[sort_index].get("direction", "asc")
            records.sort(
                key=fields_sort_key(field), reverse=direction == "desc"
            )
        if "maxRecords" in request.query:
            max_records = int(request.query["maxRecords"])
            records = records[:max_records]
        offset: Optional[str] = None
        if "offset" in request.query:
            for rec_idx, record in enumerate(records, start=1):
                if record["id"] == request.query["offset"]:
                    records = records[rec_idx:]
                    break
            else:
                raise HTTPBadRequest(reason="Wrong sort parameter5")
        page_size = int(request.query.get("pageSize", "100"))
        if len(records) > page_size:
            records = records[:page_size]
            offset = records[-1]["id"]
        if "fields[]" in request.query:
            fields = request.query.getall("fields[]")
            for record in records:
                record["fields"] = {
                    key: value
                    for key, value in record["fields"].items()
                    if key in fields
                }
        if offset is not None:
            response = {"records": records, "offset": offset}
        else:
            response = {"records": records}
        return json_response(response)

    async def retrieve_record(self, request: Request) -> Response:
        base_id = request.match_info["base_id"]
        table_name = request.match_info["table_name"]
        record_id = request.match_info["record_id"]
        if (base_id, table_name) not in self._tables:
            raise HTTPNotFound(reason="Table not found")
        for record in self._tables[base_id, table_name]:
            if record["id"] == record_id:
                return json_response(record)
        else:
            raise HTTPNotFound()

    async def create_record(self, request: Request) -> Response:
        base_id = request.match_info["base_id"]
        table_name = request.match_info["table_name"]
        fields = (await request.json())["fields"]
        created_time = datetime.now().strftime(DT_FORMAT)
        record = Record(
            id=f"record{self._loop.time()}",
            fields=fields,
            createdTime=created_time,
        )
        if (base_id, table_name) in self._tables:
            self._tables[base_id, table_name].append(record)
        else:
            self._tables[base_id, table_name] = [record]
        return json_response(record)

    async def update_record(self, request: Request) -> Response:
        base_id = request.match_info["base_id"]
        table_name = request.match_info["table_name"]
        record_id = request.match_info["record_id"]
        if (base_id, table_name) not in self._tables:
            raise HTTPNotFound(reason="Table not found")
        for record in self._tables[base_id, table_name]:
            if record["id"] == record_id:
                record["fields"] = (await request.json())["fields"]
                return json_response(record)
        else:
            raise HTTPNotFound()

    async def delete_record(self, request: Request) -> Response:
        base_id = request.match_info["base_id"]
        table_name = request.match_info["table_name"]
        record_id = request.match_info["record_id"]
        if (base_id, table_name) not in self._tables:
            raise HTTPNotFound(reason="Table not found")
        table = self._tables[base_id, table_name]
        for index, record in enumerate(table.copy()):
            if record["id"] == record_id:
                table.pop(index)
                deleted_record = DeletedRecord(id=record_id, deleted=True)
                return json_response(deleted_record)
        else:
            raise HTTPNotFound()


@pytest_asyncio.fixture
def url() -> URL:
    url = URL("https://api.airtable.com/v0/base_id/table_name")
    return url.with_query(maxRecords=0)


@pytest_asyncio.fixture
def dt_str() -> str:
    return datetime.now().strftime(DT_FORMAT)


@pytest_asyncio.fixture
async def server(dt_str: str) -> AsyncGenerator[AirtableServer, None]:
    server = AirtableServer("some_key")
    records = [
        Record(
            id=f"record{index:03d}",
            fields={
                "field_1": f"value_1_{index:03d}",
                "field_2": f"value_2_{index:03d}",
                "field_3": f"value_3_{index:03d}",
            },
            createdTime=dt_str,
        )
        for index in range(200)
    ]
    server.add_records("base_id", "table_name", records)
    await server.start()
    yield server
    await server.stop()


@pytest_asyncio.fixture
async def airtable(server: AirtableServer) -> AsyncGenerator[Airtable, None]:
    airtable = Airtable("some_key", server.connector)
    yield airtable
    await airtable.close()


def test_backoff_wait_gen() -> None:
    wait_gen = aat.backoff_wait_gen(aat.AT_WAIT)
    wait_gen.send(None)
    for value, i in zip(wait_gen, range(16)):
        assert value == aat.AT_WAIT + 2**i


@pytest.mark.asyncio
async def test_backoff() -> None:
    class TestClient:

        def __init__(self) -> None:
            self.first_attempt = True

        @backoff.on_exception(
            backoff_wait_gen,
            ClientResponseError,
            giveup=backoff_giveup,
            at_wait=0.01,
        )
        async def request(self) -> None:
            if self.first_attempt:
                self.first_attempt = False
                raise client_response_error(429)
            else:
                raise client_response_error(430)

    tc = TestClient()
    with pytest.raises(ClientResponseError) as exc_info:
        await tc.request()
    assert exc_info.type is ClientResponseError
    assert exc_info.value.status == 430
    assert not tc.first_attempt


def client_response_error(status: int) -> ClientResponseError:
    url = URL("example.com")
    info = RequestInfo(url, "GET", CIMultiDictProxy(CIMultiDict()), url)
    return ClientResponseError(info, tuple(), status=status)


@given(integers(min_value=100, max_value=526))
def test_backoff_giveup(status: int) -> None:
    backoff_flag = status not in (429, 502, 503, 504)
    assert aat.backoff_giveup(client_response_error(status)) == backoff_flag


def test_backoff_giveup_wrong_exception() -> None:
    with pytest.raises(AssertionError):
        aat.backoff_giveup(ValueError())


def test_build_repr() -> None:
    repr_str = aat.build_repr("A", b=1, c=2.34, d="efg")
    assert repr_str == "A(b=1, c=2.34, d='efg')"


@pytest.mark.asyncio
async def test_airtable_repr(airtable: Airtable) -> None:
    assert repr(airtable) == "Airtable(api_key='...')"


@pytest.mark.asyncio
async def test_airtable_client(
    server: AirtableServer, airtable: Airtable, url: URL
) -> None:
    assert isinstance(airtable.client, ClientSession)
    with pytest.raises(
        ClientResponseError, match="Authorization header absent"
    ):
        await airtable.client.get(url)
    assert server.requests() == []


@pytest.mark.asyncio
async def test_airtable_underscore_request(
    server: AirtableServer,
    airtable: Airtable,
    url: URL,
) -> None:
    assert await airtable._request("GET", url, RecordList) == RecordList(
        records=()
    )
    assert server.requests() == [RequestData("GET", url, None)]


@pytest.mark.asyncio
async def test_airtable_request(
    server: AirtableServer,
    airtable: Airtable,
    url: URL,
) -> None:
    loop = asyncio.get_running_loop()
    time1 = loop.time()
    assert await airtable.request(
        "base_id",
        "GET",
        url,
        RecordList,
    ) == RecordList(records=())
    assert server.requests() == [RequestData("GET", url, None)]
    assert await airtable.request(
        "base_id",
        "GET",
        url,
        RecordList,
    ) == RecordList(records=())
    time2 = loop.time()
    assert time2 - time1 >= aat.AT_INTERVAL
    assert server.requests() == [
        RequestData("GET", url, None),
        RequestData("GET", url, None),
    ]


@pytest.mark.asyncio
async def test_airtable_close() -> None:
    airtable = Airtable("secret_key")
    await airtable.close()
    assert airtable._client.closed


@pytest.mark.asyncio
async def test_airtable_context(airtable: Airtable) -> None:
    async with airtable:
        pass
    assert airtable._client.closed


@pytest.mark.asyncio
async def test_airtable_base(airtable: Airtable) -> None:
    base = airtable.base("some_base_id")
    assert isinstance(base, AirtableBase)
    assert base.id == "some_base_id"


@pytest.mark.asyncio
async def test_airtable_base_repr(airtable: Airtable) -> None:
    base = airtable.base("some_base_id")
    assert repr(base) == (
        "AirtableBase(base_id='some_base_id', "
        "airtable=Airtable(api_key='...'))"
    )


@pytest.mark.asyncio
async def test_airtable_base_id(airtable: Airtable) -> None:
    base = airtable.base("some_base_id")
    assert base.id == "some_base_id"


@pytest.mark.asyncio
async def test_airtable_base_url(airtable: Airtable) -> None:
    base = airtable.base("some_base_id")
    assert base.url == aat.API_URL / "some_base_id"


@pytest.mark.asyncio
async def test_airtable_base_request(
    server: AirtableServer, airtable: Airtable, url: URL
) -> None:
    base = airtable.base("base_id")
    assert await base.request("GET", url, RecordList) == RecordList(records=())
    assert server.requests() == [RequestData("GET", url, None)]


class S(Struct):
    pass


@pytest.mark.asyncio
async def test_airtable_base_table(airtable: Airtable) -> None:
    base = airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    assert isinstance(table, AirtableTable)
    assert table.name == "some_table"
    assert table.url == base.url / "some_table"
    assert table.base == base


@pytest.mark.asyncio
async def test_airtable_table_request(
    server: AirtableServer,
    airtable: Airtable,
    url: URL,
) -> None:
    base = airtable.base("base_id")
    table = base.table("table_name", Struct)
    assert await table._request("GET", url, RecordList) == RecordList(
        records=()
    )
    assert server.requests() == [RequestData("GET", url, None)]


class F(Struct):
    field_1: str | None = None
    field_2: str | None = None
    field_3: str | None = None


@pytest.mark.asyncio
async def test_airtable_table_list_records(
    server: AirtableServer,
    airtable: Airtable,
    dt_str: str,
) -> None:
    base = airtable.base("base_id")
    table = base.table("table_name", F)
    records, offset = await table.list_records(
        fields=("field_1", "field_2"),
        filter_by_formula="{field_3}",
        max_records=100500,
        page_size=3,
        sort=(("field_1", SortDirection.ASC), ("field_2", SortDirection.DESC)),
        view="table3",
        cell_format=CellFormat.JSON,
        time_zone="Europe/Moscow",
        user_locale="ru",
        offset="record033",
    )
    assert server.requests() == [
        RequestData(
            "GET",
            table.url.with_query(
                (
                    ("fields[]", "field_1"),
                    ("fields[]", "field_2"),
                    ("filterByFormula", "{field_3}"),
                    ("maxRecords", "100500"),
                    ("pageSize", "3"),
                    ("sort[0][field]", "field_1"),
                    ("sort[0][direction]", "asc"),
                    ("sort[1][field]", "field_2"),
                    ("sort[1][direction]", "desc"),
                    ("view", "table3"),
                    ("cellFormat", "json"),
                    ("timeZone", "Europe/Moscow"),
                    ("userLocale", "ru"),
                    ("offset", "record033"),
                )
            ),
            None,
        )
    ]
    assert len(records) == 3
    for index, record in enumerate(records, start=34):
        assert isinstance(record, AirtableRecord)
        assert record.id == f"record{index:03d}"
        assert record.fields.field_1 == f"value_1_{index:03d}"
        assert record.fields.field_2 == f"value_2_{index:03d}"
        assert record.fields.field_3 is None
        assert record.created_time == parse_dt(dt_str)
        assert record.table == table
    assert offset == "record036"


@pytest.mark.asyncio
async def test_airtable_table_iter_records(
    server: AirtableServer,
    airtable: Airtable,
    dt_str: str,
) -> None:
    base = airtable.base("base_id")
    table = base.table("table_name", F)
    records = [
        record
        async for record in table.iter_records(
            fields=["field_1"],
            max_records=6,
            page_size=3,
        )
    ]
    assert len(records) == 6
    for index, record in enumerate(records):
        isinstance(record, AirtableRecord)
        assert record.id == f"record{index:03d}"
        assert record.fields.field_1 == f"value_1_{index:03d}"
        assert record.fields.field_2 is None
        assert record.fields.field_3 is None
        assert record.created_time == parse_dt(dt_str)
        assert record.table == table
    assert server.requests() == [
        RequestData(
            "GET",
            table.url.with_query(
                (
                    ("fields[]", "field_1"),
                    ("maxRecords", "6"),
                    ("pageSize", "3"),
                )
            ),
            None,
        ),
        RequestData(
            "GET",
            table.url.with_query(
                (
                    ("fields[]", "field_1"),
                    ("maxRecords", "6"),
                    ("pageSize", "3"),
                    ("offset", "record002"),
                )
            ),
            None,
        ),
    ]


@pytest.mark.asyncio
async def test_airtable_table_retrieve_record(
    server: AirtableServer,
    airtable: Airtable,
    dt_str: str,
) -> None:
    base = airtable.base("base_id")
    table = base.table("table_name", F)
    record = await table.retrieve_record("record003")
    assert isinstance(record, AirtableRecord)
    assert record.id == "record003"
    assert record.fields == F(
        field_1="value_1_003",
        field_2="value_2_003",
        field_3="value_3_003",
    )
    assert record.created_time == parse_dt(dt_str)
    assert record.table == table
    assert server.requests() == [
        RequestData(
            "GET",
            table.url / "record003",
            None,
        )
    ]


@pytest.mark.asyncio
async def test_airtable_table_create_record(
    server: AirtableServer,
    airtable: Airtable,
    dt_str: str,
) -> None:
    base = airtable.base("base_id")
    table = base.table("table_name", F)
    record = await table.create_record(
        F(
            field_1="value_1_new_001",
            field_2="value_2_new_001",
            field_3="value_3_new_001",
        )
    )
    assert isinstance(record, AirtableRecord)
    assert isinstance(record.id, str)
    assert record.fields == F(
        field_1="value_1_new_001",
        field_2="value_2_new_001",
        field_3="value_3_new_001",
    )
    assert record.created_time == parse_dt(dt_str)
    assert record.table == table
    assert server.requests() == [
        RequestData(
            "POST",
            table.url,
            {
                "fields": {
                    "field_1": "value_1_new_001",
                    "field_2": "value_2_new_001",
                    "field_3": "value_3_new_001",
                }
            },
        )
    ]


@pytest.mark.asyncio
async def test_airtable_record_init(
    airtable: Airtable,
    dt_str: str,
) -> None:
    base = airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    record = AirtableRecord(
        "record1",
        Struct(),
        datetime.now(UTC),
        table,
    )
    assert isinstance(record, AirtableRecord)


@pytest.mark.asyncio
async def test_airtable_record_repr(airtable: Airtable) -> None:
    base = airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    time_string = "2021-01-25T17:28:21.000Z"
    record = AirtableRecord(
        "record1",
        Struct(),
        parse_dt(time_string),
        table,
    )
    assert repr(record) == (
        "AirtableRecord(record_id='record1', fields=Struct(), "
        "created_time=datetime.datetime(2021, 1, 25, 17, "
        "28, 21, tzinfo=datetime.timezone.utc), "
        "table=AirtableTable(table_name='some_table', "
        "base=AirtableBase(base_id='some_base_id', "
        "airtable=Airtable(api_key='...'))))"
    )


@pytest.mark.asyncio
async def test_airtable_record_table(airtable: Airtable) -> None:
    base = airtable.base("some_base_id")
    table = base.table("some_table", Struct)
    record = AirtableRecord(
        "record1",
        Struct(),
        datetime.now(UTC),
        table,
    )
    assert record.table == table


@pytest.mark.asyncio
async def test_airtable_record_request(
    server: AirtableServer,
    airtable: Airtable,
    url: URL,
    dt_str: str,
) -> None:
    base = airtable.base("base_id")
    table = base.table("table_name", F)
    f = F(
        field_1="value_1_000",
        field_2="value_2_000",
        field_3="value_3_000",
    )
    record = AirtableRecord(
        "record000",
        f,
        parse_dt(dt_str),
        table,
    )
    assert await record._request(
        "GET",
        record.url,
        aat.Record[F],
    ) == aat.Record(
        id="record000",
        fields=f,
        created_time=parse_dt(dt_str),
    )
    assert server.requests() == [RequestData("GET", record.url, None)]


@pytest.mark.asyncio
async def test_airtable_record_update(
    server: AirtableServer,
    airtable: Airtable,
    dt_str: str,
) -> None:
    base = airtable.base("base_id")
    table = base.table("table_name", F)
    f1 = F(
        field_1="value_1_000",
        field_2="value_2_000",
        field_3="value_3_000",
    )
    record = AirtableRecord(
        "record000",
        f1,
        parse_dt(dt_str),
        table,
    )
    f2 = F(
        field_1="value_1_new_000",
        field_2="value_2_new_000",
        field_3="value_3_new_000",
    )
    await record.update(f2)
    assert record.fields == f2
    assert server.requests() == [
        RequestData(
            "PATCH",
            record.url,
            {
                "fields": {
                    "field_1": "value_1_new_000",
                    "field_2": "value_2_new_000",
                    "field_3": "value_3_new_000",
                }
            },
        )
    ]


@pytest.mark.asyncio
async def test_airtable_record_delete(
    server: AirtableServer,
    airtable: Airtable,
    dt_str: str,
) -> None:
    base = airtable.base("base_id")
    table = base.table("table_name", Struct)
    record = AirtableRecord(
        "record000",
        Struct(),
        datetime.now(UTC),
        table,
    )
    await record.delete()
    assert record.deleted
    assert server.requests() == [RequestData("DELETE", record.url, None)]
    with pytest.raises(RuntimeError, match="Record is already deleted"):
        await record.delete()
    assert server.requests() == [RequestData("DELETE", record.url, None)]
    with pytest.raises(RuntimeError, match="Record is deleted"):
        await record.update(Struct())
    assert server.requests() == [RequestData("DELETE", record.url, None)]



__version__ = "0.1.4"

from .aioairtable import (
    Airtable,
    AirtableBase,
    AirtableRecord,
    AirtableTable,
    Attachment,
    CellFormat,
    Collaborator,
    Method,
    NewAttachment,
    SortDirection,
    Thumbnail,
)

__all__ = (
    "Airtable",
    "AirtableBase",
    "AirtableRecord",
    "AirtableTable",
    "Attachment",
    "CellFormat",
    "Collaborator",
    "Method",
    "NewAttachment",
    "SortDirection",
    "Thumbnail",
    "__version__",
)
